from .node import Node
import logging
import threading

logging.basicConfig(level=logging.INFO)

log = logging.getLogger("graph")


class Graph:
    def __init__(self) -> None:
        self.nodes: dict[str, Node] = {}
        self.num_starter_nodes = 0

    def from_json(self, input: dict) -> None:
        # Create all nodes without dependencies.
        nodes = input["nodes"]
        for name, node in nodes.items():
            n = Node(name)
            if "dependencies" not in node:
                self.num_starter_nodes += 1
            self.nodes[name] = n

        # Create an init node from which the graph starts.
        self.nodes["__init"] = Node("__init")

        # Create all dependencies.
        for node in self.nodes.values():
            if node.name == "__init":
                continue
            if "dependencies" not in nodes[node.name]:
                node.dependencies.append(self.nodes["__init"])
                continue
            for dep in nodes[node.name]["dependencies"]:
                node.dependencies.append(self.nodes[dep])
                self.nodes[dep].num_downstream += 1

    def run(self) -> None:
        # Run the graph.
        threads = []
        for node in self.nodes.values():
            if node.name == "__init":
                continue
            t = threading.Thread(target=node.listen, args=())
            t.start()
            threads.append(t)

        # Send enough sparks to start the graph.
        for _ in range(self.num_starter_nodes):
            self.nodes["__init"].output_queue.put([{"spark": "ignited"}])
        log.info("Sent spark to %d nodes", self.num_starter_nodes)
        for t in threads:
            t.join()
        log.info("Graph finished running.")



import queue
from typing import Any
import logging

log = logging.getLogger("node")


class Node:
    def __init__(self, name) -> None:
        self.name: str = name
        self.output_queue: queue.Queue = queue.Queue()
        self.dependencies: list[Node] = []
        self.num_downstream: int = 0
        self.state: dict[str, Any] = {}

    def run(self, dep_inputs: dict[str, list[dict[str, Any]]]) -> None:
        for i in range(self.num_downstream):
            self.output_queue.put(
                [
                    {
                        self.name: self.state,
                    }
                ]
            )

    def listen(self) -> None:
        dep_outputs: dict[str, list[dict[str, Any]]] = {}
        for dep in self.dependencies:
            dep_outputs[dep.name] = dep.output_queue.get()
        log.info(f"Running node {self.name}  with received {len(dep_outputs)} dependencies")
        self.run(dep_outputs)
